/*
  File name: I3Dg.h
  Copyright: (C) 2003 - 2007 Tomasz Kazmierczak

  Creation date: 18.10.2003
  Last modification date: 23.09.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#ifndef _I3DG_H_
#define _I3DG_H_

#include <stdint.h>
#include <map>

typedef uint32_t MeshBase3DgUID;
typedef uint32_t Mesh3DgUID;

//errors
enum Error3Dg {
	ERR3DG_OK,
	ERR3DG_OUTOFMEMORY,
	ERR3DG_INVALIDCALL,
	ERR3DG_INVALIDARG
};

//light types
typedef enum LightType3Dg {
	POINT_LIGHT = 1,
	SPOT_LIGHT = 2,
	DIRECTIONAL_LIGHT = 3,
	LIGHT_FORCE_DWORD = 0xFFFFFFFF
};

//texture types
typedef enum TextureType3Dg {
	UNKNOWN_TEX_TYPE = 0,
	PIXEL_MAP = 1,              //24 bpp or 32 bpp
	BUMP_MAP = 2,               //8 bpp
	LIGHT_MAP = 3,              //24 bpp
	ALPHA_MAP = 4,              //8 bpp
	ENVIRONMENT_MAP = 5,        //24 bpp
	TEX_TYPE_FORCE_DWORD = 0xFFFFFFFF
};

//texture wrap modes
typedef enum TextureWrapMode3Dg {
	TEX_WRAP_REPEAT = 1,
	TEX_WRAP_MIRRORED_REPEAT,
	TEX_WRAP_CLAMP,
	TEX_WRAP_CLAMP_TO_EDGE,
	TEX_WRAP_CLAMP_TO_BORDER,
	TEX_WRAP_FORCE_DWORD = 0xFFFFFFFF
};

class ColorRGBA3Dg {
public:
	float r;
	float g;
	float b;
	float a;
};

class Object3Dg {
protected:
	Object3Dg *mParent;
	Object3Dg *mChild;
public:
	Object3Dg() : mParent(0), mChild(0) {};

	void setParent(Object3Dg *parent) { mParent = parent; };
	void setChild(Object3Dg *child) { mChild = child; };
	Object3Dg *getParent() { return mParent; };
	Object3Dg *getChild() { return mChild; };
};

class Material3Dg {
private:
	int32_t    *pMtlTextures;

public:
	ColorRGBA3Dg sDiffuse;
	ColorRGBA3Dg sAmbient;
	ColorRGBA3Dg sSpecular;
	ColorRGBA3Dg sEmissive;
	float        fShininess;
	uint32_t     uiTexturesQuantity;

	Material3Dg();
	~Material3Dg();
	bool addTexture(int32_t iTexNum);
	//returns an identifier of uiMtlTexNum'th material's texture, or -1 if there isn't such
	int32_t getTexture(uint32_t uiMtlTexNum);
};

class Texture3Dg {
public:
	TextureType3Dg      eType;
	TextureWrapMode3Dg  eSWrapMode;
	TextureWrapMode3Dg  eTWrapMode;
	uint16_t            uiPathBufferSize;
	char               *pPath;
};

typedef class Point3Dg {
public:
	Point3Dg();
	Point3Dg(float x, float y, float z);
	union {
		class {
		public:
			float x;
			float y;
			float z;
		};
		float pPt[3];
	};
};

typedef class {
public:
	union {
		class {
		public:
			float a;
			float b;
			float c;
			float d;
		};
		float pPl[4];
	};
} Plane3Dg;

typedef class Vector3Dg : public Point3Dg {
public:
	Vector3Dg() : Point3Dg() {};
	Vector3Dg(float x, float y, float z) : Point3Dg(x, y, z) {};
	Vector3Dg(Point3Dg &point) : Point3Dg(point) {};

	void operator +=(Vector3Dg &vec);
	void operator -=(Vector3Dg &vec);
	void operator *=(float multiplier);
	void operator /=(float divisor);
	Vector3Dg operator *(float multiplier);
	Vector3Dg operator +(const Vector3Dg &vec);
	Vector3Dg operator -(const Vector3Dg &vec);
	float dotProduct(const Vector3Dg &vec);
	Vector3Dg crossProduct(const Vector3Dg &vec);
	void normalize();
	float length();
	float lengthPow2();
};

class Vertex3Dg {
public:
	Point3Dg  sCoords;
	Point3Dg  sNormal;
	float     fTex[2];
};

class Triangle3Dg {
public:
	Vertex3Dg sVertexA;
	Vertex3Dg sVertexB;
	Vertex3Dg sVertexC;
};

class Matrix3Dg {
public:
	union {
		class {
		public:
			float R1C1, R1C2, R1C3, R1C4;
			float R2C1, R2C2, R2C3, R2C4;
			float R3C1, R3C2, R3C3, R3C4;
			float R4C1, R4C2, R4C3, R4C4;
		};
		float pM[4][4];
	};
};

class MeshBase3Dg : public Object3Dg {
private:
	//unique identifier
	MeshBase3DgUID mUID;
	Triangle3Dg   *mTriangles;
	uint32_t       mTriansQuantity;

public:

	MeshBase3Dg(MeshBase3DgUID uid);
	~MeshBase3Dg();

	bool addTrian(Triangle3Dg *triangle);
	void releaseTriangles();

	uint32_t getTriansQuantity() { return mTriansQuantity; };
	Triangle3Dg *getTriangles() { return mTriangles; };
	Triangle3Dg *getTriangle(uint32_t trianglenumber);
	MeshBase3DgUID getUID() { return mUID; };
};

class MeshAnim3Dg {
public:
	Point3Dg   sTranslation;
	Point3Dg   sRotation;
	Point3Dg   sScaling;
	int32_t    iElapsedTime;
	int32_t    iAnimStatus;
	//if this variable is non-zero then there is a geometry animation
	uint32_t   dwGeomAnimTime;
	//if this variable is non-zero then there is a texture animation
	uint32_t   dwTexAnimTime;
	uint32_t   dwBeginMtl;
	uint32_t   dwNumMtls;
};

enum {
	MRP_HAS_BASE =       1 << 0,
	MRP_RENDERABLE =     1 << 1,
	MRP_TRANSPARENT =    1 << 2,
	MRP_Z_PLUS_OFFSET =  1 << 3,
	MRP_Z_MINUS_OFFSET = 1 << 4,
	MRP_Z_WRITE =        1 << 5,
};

class Mesh3Dg : public Object3Dg {
private:
	//mesh's base
	MeshBase3Dg *mBase;
	//properties
	uint32_t mProps;
	//unique identifier
	Mesh3DgUID mUID;

public:
	//mesh's transformations
	Matrix3Dg     mRotation;
	Matrix3Dg     mTranslation;
	Matrix3Dg     mScaling;

	//mesh's animation
	MeshAnim3Dg   mAnimation;
	//material number
	int32_t       mMaterialNumber;

	Mesh3Dg(Mesh3DgUID uid);
	~Mesh3Dg();

	void rotate(float x, float y, float z);
	void setPosition(float x, float y, float z);
	void addTranslation(float x, float y, float z);
	void scale(float x, float y, float z);

	Vector3Dg geomCenter();

	//set

	//mesh's base (returns pointer to previous, or 0 if there was no previous)
	MeshBase3Dg *setBase(MeshBase3Dg *base);
	//rendering properties
	void setProperty(uint32_t property, bool set);
	void setProps(uint32_t props) { mProps = props; };

	//get

	//mesh's base
	MeshBase3Dg *getBase() { return mBase; };
	//rendering properties
	bool getProperty(uint32_t property) { return mProps & property; };
	uint32_t getProps() { return mProps; };

	Mesh3DgUID getUID() { return mUID; };
};

class Light3Dg {
public:
	LightType3Dg eLightType;
	ColorRGBA3Dg sDiffuse;
	ColorRGBA3Dg sSpecular;
	ColorRGBA3Dg sAmbient;
	Point3Dg     sPosition;
	Point3Dg     sDirection;
	float        fRange;
	float        fFalloff;
	float        fAttenuation0;
	float        fAttenuation1;
	float        fAttenuation2;
	float        fTheta;
	float        fPhi;
};

typedef std::map<Mesh3DgUID, Mesh3Dg *> MeshMap3Dg;
typedef std::map<MeshBase3DgUID, MeshBase3Dg *> MeshBaseMap3Dg;

//3Dgeom interface class
class I3Dg : public Object3Dg {
protected:
	Error3Dg    mError;

	uint32_t  mLightsQuantity;
	uint32_t  mTexturesQuantity;
	uint32_t  mMaterialsQuantity;

public:
	MeshBaseMap3Dg  mMeshBases;
	MeshMap3Dg      mMeshes;
	Light3Dg      **mLights;
	Texture3Dg    **mTextures;
	Material3Dg   **mMaterials;

	I3Dg();
	virtual ~I3Dg();
	virtual void release();

	MeshBase3Dg *createMeshBase(MeshBase3DgUID uiMBUID);
	Mesh3Dg     *createMesh(Mesh3DgUID uiMUID);
	Light3Dg    *createLight();
	Material3Dg *createMaterial();
	Texture3Dg  *createTexture(char *pTexPath, uint16_t uiPathBufferSize);

	MeshBase3DgUID generateMeshBaseUID();
	Mesh3DgUID generateMeshUID();

	void destroyMesh(Mesh3DgUID uiMUID);

	uint32_t getMaterialsQuantity() { return mMaterialsQuantity; };
	uint32_t getTexturesQuantity() { return mTexturesQuantity; };
	uint32_t getLightsQuantity() { return mLightsQuantity; };

	Error3Dg getLastError() { return mError; };
	const char *getLastErrorDesc();
	const char *getErrorDesc(Error3Dg eError);
};

I3Dg *create3DgInterface();

#endif //_I3DG_H_
